
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unordered_set>
#include <vector>
#include "ftr.hh"
#include "makecfg.hh"

#include "u64.hh"

static const bool ZIRCONIUM_COSTS = true;
static const bool SUPPORT_LITERALS = false;
static const bool LOCAL_STACK_ONLY = true;
static const unsigned COUNT_INTERVAL = 0x1ffff;
static const unsigned LEADOUT_SIZE = 3 ;
static const ulongint LEADOUT_PC = 0x100;
static const unsigned BRANCH_COST = 1 ;
static const unsigned STACK_ITEM_SIZE = 8 ;    
static const unsigned FP_REGISTER = 15 ;    
static const unsigned GP_REGISTER = 29 ;    
static const unsigned SP_REGISTER = 30 ;    
static const unsigned ZERO_REGISTER = 31 ;    
static const unsigned NOT_A_REGISTER = 99 ;    
static const unsigned RFILE_SIZE = 100 ;    
static const unsigned MAX_OFFSET = 1 << 28 ;     /* bytes */
static const unsigned MAX_STACK_OFFSET = 1 << 24 ;     /* bytes */

#define BASE_EDGE_ATTR "a(\"EDGECOLOR\",\"#ff0000\"),"

enum AP_Class { AP_NOT_ITERATIVE, AP_CONSTANT_STEP,
        AP_VARIABLE_STEP, AP_RANDOM_ACCESS, AP_PSEUDO_STATIC };

struct Access_Data {
    ulongint        counter;
    AP_Class        ap_classification;
    ulongint        sample;
    int             ap_constant_step, last_offset, initial_offset;
};


// Base_Creator: Represents an instruction that creates a new base pointer
class Base_Creator
{
public:
    Base_Creator(ulongint _pc, ulongint _id, bool _is_stack_pointer);
    virtual ~Base_Creator() {};
    void Export ( FILE * outfd ) ;
    void Export_UDG ( FILE * udgf ) ;
    bool Is_SP() { return is_stack_pointer; }
    bool Is_Valid() { return (upper_bound != INT_MIN); }
    bool Is_Important() { return (!Is_SP()) && Is_Valid(); }

    ulongint        pc;
    ulongint        id;
    ulongint        max_instances;
   
    bool            is_stack_pointer, udg_printed;
    int32_t         upper_bound, lower_bound;
    std::set<ulongint>  active_instances;
    std::set<ulongint>  inactive_instances;
    std::map<ulongint, Access_Data>   ap_last_access;
} ;

// Base_Instance: Tracks a base pointer through the register file & stack
class Base_Instance
{
public:
    Base_Instance(Base_Creator * _source, ulongint _initial_value);
    virtual ~Base_Instance();
    bool Is_SP() { return source->Is_SP(); }
    int32_t Stack_Offset(ulongint ea_value);
    void Use(ulongint address,  ulongint pc);

    void Inc_Ref_Count() { ref_count ++; }
    ulongint Dec_Ref_Count();
    
    Base_Creator *  source;
    ulongint        initial_value;
    ulongint        instance_number;
    ulongint        ref_count;

private:
    Base_Instance(const Base_Instance &x) { assert(0); }
    Base_Instance & operator= (const Base_Instance &x) { assert(0); }
} ;

void BP_Destroy(Base_Instance * bi);
   
typedef struct Usage_Data_struct {
        Base_Creator *      source;
        std::set<ulongint>  used_at;
    } Usage_Data;

class CFG_Edge
{
public:
    CFG_Edge ( ulongint id , CFG_Vertex * start ,
                CFG_Vertex * end , ulongint frequency ) ;
    virtual ~CFG_Edge () {} ;
    void Export ( FILE * outfd ,
        std::unordered_set<ulongint> & mem_access ) ;
    void Add_Instruction ( ulongint pc , ulongint repeats , ulongint cost ) ;
    bool Merge ( CFG_Edge * other ) ;
    void Check_Instruction ( std::unordered_set<ulongint> & known ) ;

    Base_Instance * Define(ulongint _pc, ulongint _id,
                ulongint _initial_value, bool _is_stack_pointer);
    void Use (Base_Instance * bi, ulongint pc, ulongint address);
    ulongint Start_PC();
    ulongint End_PC();

    ulongint                frequency , id , branch_cost ;
    CFG_Vertex *            start ;
    CFG_Vertex *            end ;
    std::vector<std::pair<ulongint, ulongint> > path ;
    BType                   branch_type;
    std::map<ulongint, Base_Creator *>  creator; // key is pc
    std::map<ulongint, Usage_Data>      user;    // key is pc
} ;


class CFG_Edge_Set : public std::set<CFG_Edge *> {} ;


class CFG_Vertex
{
public:
    CFG_Vertex ( ulongint id , ulongint pc ) ;
    virtual ~CFG_Vertex () {} ;
    bool Trim () ;
    void Export ( FILE * outfd ) ;
    void Check_Frequency () ;
    void Add_Edges ( std::list<CFG_Edge *> & all_edges ) ;
    void Export_UDG ( FILE * udgf , Symbol_Table * symtab ) ;
    void Present ( Base_Instance * bi );

    bool Is_Entrance () { return entrances . size () == 0 ; } ;
    bool Is_Exit () { return exits . size () == 0 ; } ;

    CFG_Edge_Set        entrances , exits ;
    ulongint            pc , id ;
    bool                udg_printed ;
    std::map<ulongint, Base_Creator *> present;     // key is pc
} ;


Make_CFG :: Make_CFG ()
{
    entrance = 0L ;
    uid = 0;
}

Make_CFG :: ~Make_CFG ()
{
    Unmake () ;
}

void Make_CFG :: Read_Trace ( FILE * ftr0fd ,
            ulongint main_start , ulongint main_end ,
            FILE * exe_trace_out )
{
    struct FastTraceRecord      ftrp [ 1 ] ;
    struct FastTraceRecord *    ftr0 = & ftrp [ 0 ] ;
    bool                        about_to_activate = false ;
    bool                        about_to_deactivate = false ;
    bool                        active = false ;
    bool                        can_activate = true ;
    bool                        unknown_warn = false ;
    ulongint                    expect_pc = 0 ;
    ulongint                    current_pc = 0 ;
    ulongint                    source_pc = 0 ;
    ulongint                    num_repeats = 1 ;
    ulongint                    counter = 0 ;
    ulongint                    total_cost = 0;
    ulongint                    i , data_available = LEADOUT_SIZE;
    BType                       old_bt = BT_None ;
    struct FastTraceRecord      leadout [ LEADOUT_SIZE ] ;

    memset ( leadout , 0 , sizeof ( leadout ) ) ;
    for ( i = 0 ; i < LEADOUT_SIZE ; i ++ )
    {
        leadout [ LEADOUT_SIZE - ( 1 + i ) ] . optype = 'A';
        leadout [ LEADOUT_SIZE - ( 1 + i ) ] . cost = 1;
        leadout [ LEADOUT_SIZE - ( 1 + i ) ] . pc_value = 
            LEADOUT_PC + ( i * 4 ) ;
    }
    leadout [ 0 ] . optype = 'X';

    bb_mark . clear () ;
    fall_through . clear () ;
    exec_frequency . clear () ;
    branch_links . clear () ;
    branch_type . clear () ;
    cost_lookup . clear () ;
    repeat_lookup . clear () ;
    mem_access . clear () ;
    uid = 0;
    inst_count = 0;
    early_exit_flag = false;

    while ( data_available )
    {
        if ( fread ( ftr0 , 
            sizeof ( struct FastTraceRecord ) , 1 , ftr0fd ) == 1 )
        {
            data_available = LEADOUT_SIZE;
        } else {
            if ( active )  
            {
                // Entered leadout while active => early exit!
                early_exit_flag = true;
            }
            data_available --;
            memcpy ( ftr0 , & leadout [ data_available ] ,
                    sizeof ( struct FastTraceRecord ) ) ;
        }

        bool is_trap = ( ftr0 -> optype == 'T' ) ;
        BType bt = BT_None ;
        current_pc = ftr0 -> pc_value ;
        assert (( current_pc & 3 ) == 0 ) ;
        source_pc = expect_pc - 4 ;
        counter ++;

        if ( is_trap )
        {
            continue ;
        }

        //fprintf ( stderr , "%"FX" is %c\n" , current_pc , ftr0 -> optype ) ;
        switch ( ftr0 -> optype )
        {
            case 'X' :      
                    bt = branch_type [ current_pc ] = BT_Serialising ;
                    break ;
            case 'B' :      
                    bt = branch_type [ current_pc ] = BT_Conditional ;
                    break ;
            case 'J' :      
                    bt = BT_Unconditional ;
                    break ;
            case 'L' :
            case 'S' :
                    mem_access . insert ( current_pc ) ;
                    break;
            case '?' :
                    if ( ! unknown_warn )
                    {
                        unknown_warn = true;
                        fprintf ( stderr , 
                            "Warning: Unknown integer ALU op at 0x%"FX"\n" ,
                            current_pc ) ;
                    }
                    break ;
            default :       
                    break ;
        }
        if ( ZIRCONIUM_COSTS )
        {
            ftr0 -> cost = 1;
        }

        // Keep track of costs, and the repeat counter, which is
        // very important for LDM/STM operations on ARM.
        if ( cost_lookup . count ( current_pc ) == 0 )
        {
            cost_lookup [ current_pc ] = ftr0 -> cost ;
            assert ( ftr0 -> cost != 0 );
        } else {
            assert ( cost_lookup [ current_pc ] == ftr0 -> cost );
        }
        if ( repeat_lookup . count ( current_pc ) == 0 )
        {
            repeat_lookup [ current_pc ] = num_repeats ;
        } 
        if ((counter & COUNT_INTERVAL) == 0) {
            fprintf (stderr, "Read_Trace %"FU" %s\n", counter, 
                active ? "active" : 
                    ( can_activate ? "before" : "after" ));
        }

        // Activation mechanism - activation begins when
        // "main" calls "wcetmain", and ends when "wcetmain" returns.
        bool in_main = (( main_start <= current_pc )
            && ( current_pc < main_end )) ;
        if ( active )
        {
            if ( in_main || ( data_available != LEADOUT_SIZE )) 
            {
                // deactivate
                bb_mark . insert ( source_pc ) ;
                about_to_deactivate = true ;
            }
            about_to_activate = false ;
            can_activate = false ;
        } else {
            if ( in_main ) 
            {
                about_to_activate = can_activate ;
            } else if ( about_to_activate )
            {
                active = true ;
                bb_mark . insert ( current_pc ) ;
            }
        }

        if ( ! active )
        {
            continue ;
        }

        total_cost += ftr0 -> cost;

        if (( exe_trace_out != 0L )
        && ( current_pc != source_pc ))
        {
            fprintf ( exe_trace_out , "%"FX"\n" , current_pc ) ;
        }

        if ( current_pc == source_pc )
        {
            // Repeat instruction (e.g. stm/ldm)
            num_repeats ++ ;
            if ( repeat_lookup [ current_pc ] < num_repeats )
            {
                repeat_lookup [ current_pc ] = num_repeats ;
            }
        } else if (( current_pc != expect_pc )
        || ( old_bt == BT_Unconditional ))
        {
            // Branch to new basic block
            num_repeats = 1 ;
            bb_mark . insert ( expect_pc ) ;
            bb_mark . insert ( current_pc ) ;
            total_cost += BRANCH_COST;

            // Record branch link
            if ( branch_links . count ( source_pc ) != 0 )
            {
                // Branch source has been seen before
                if ( branch_links [ source_pc ] . count (
                                    current_pc ) != 0 )
                {
                    // This branch has been seen before
                    branch_links [ source_pc ][ 
                                current_pc ] ++ ;
                } else {
                    // New branch destination
                    branch_links [ source_pc ][ 
                                current_pc ] = 1 ;
                }
            } else {
                // New branch source and destination
                branch_links [ source_pc ][ 
                            current_pc ] = 1 ;
            }
        } else {
            // Fall through
            num_repeats = 1 ;
            if ( old_bt != BT_None )
            {
                // There's a branch here as well
                bb_mark . insert ( current_pc ) ;
            }
            if ( fall_through . count ( source_pc ) == 0 )
            {
                fall_through [ source_pc ] = 1 ;
            } else {
                fall_through [ source_pc ] ++ ;
            }
        }

        if ( current_pc != source_pc )
        {
            if ( exec_frequency . count ( source_pc ) == 0 )
            {
                exec_frequency [ source_pc ] = 1 ;
            } else {
                exec_frequency [ source_pc ] ++ ;
            }
        }

        if ( about_to_deactivate )
        {
            active = false ;
        }
        expect_pc = current_pc + 4 ;
        old_bt = bt ;
    }
    if ( can_activate || active )
    {
        // Before main, or during main - shouldn't exit here.
        // Can't give an exact location due to buffering in the FIFO.
        fprintf ( stderr , "Warning: early exit after PC 0x%"FX"\n" ,
                current_pc ) ;
        // But we *do* exit early if --maxtick is being used so must
        // handle this correctly
    }
    fprintf ( stderr , "Measured cost = %"FU"\n", total_cost ) ;
    inst_count = counter;

    // The end
    bb_mark . erase ( 0 ) ;
}


void Make_CFG :: Make ()
{
    std::set<ulongint>::iterator            i ;
    std::map<ulongint, ulongint>::iterator  k ;
    std::map<ulongint, CFG_Vertex *>        cfg_vertex ;
    std::list<CFG_Vertex *>::iterator       l ;
    ulongint                                bb_start , bb_end , j ;

    Unmake () ;
    // Step 1 - create CFG vertexes
    for ( i = bb_mark . begin () ; i != bb_mark . end () ; i ++ )
    {
        CFG_Vertex * cv = new CFG_Vertex ( uid , (* i) ) ;
        cfg_vertex [ (* i) ] = cv ;
        all_cfg_vertexes . push_back ( cv ) ;
        uid ++ ;
    }
    // Step 2 - create CFG edges
    i = bb_mark . begin () ;
    bb_start = (* i) ;
    i ++ ;
    for ( ; i != bb_mark . end () ; i ++ )
    {
        // Get data about this basic block
        bb_end = (* i) - 4 ;
        ulongint frequency = 0 ;

        if ( exec_frequency . count ( bb_start ) == 1 )
        {
            frequency = exec_frequency [ bb_start ] ;
        }
        
        if (( frequency == 0 )
        && ( branch_links . count ( bb_end ) == 0 ))
        {
            // Never executed - dead code/gap between functions
            bb_start = (* i) ;
            continue ;
        }

        // Lookup/make vertexes
        CFG_Vertex * cv_start = cfg_vertex [ bb_start ] ;
        CFG_Vertex * cv_end = new CFG_Vertex ( uid , bb_end ) ;
        all_cfg_vertexes . push_back ( cv_end ) ;
        uid ++ ;

        // Add edge to represent basic block
        CFG_Edge * bb = new CFG_Edge ( uid , cv_start , cv_end , frequency);
        uid ++ ;

        // By definition - no branches in the middle of a basic block!
        for ( j = bb_start ; j < bb_end ; j += 4 )
        {
            assert ( branch_links . count ( j ) == 0 ) ;
            assert ( exec_frequency . count ( j ) == 1 ) ;
            // Check that execution frequencies are constant within
            // each basic block
            assert ( exec_frequency [ j ] == frequency ) ;
            assert ( frequency != 0 ) ;
            bb -> Add_Instruction ( j , 
                        repeat_lookup [ j ] , cost_lookup [ j ] ) ;
        }
        bb -> Add_Instruction ( bb_end , 
                repeat_lookup [ bb_end ] , cost_lookup [ bb_end ] ) ;
        
        // Branch type (if any)
        if ( branch_type . count ( bb_end ) != 0 )
        {
            bb -> branch_type = branch_type [ bb_end ] ;
        } else {
            bb -> branch_type = BT_Unconditional ;
        }

        // At end of basic block - maybe fall-through to next 
        if ( fall_through . count ( bb_end ) == 1 )
        {
            ulongint bb_target = bb_end + 4 ;
            ulongint frequency = fall_through [ bb_end ] ;

            assert ( frequency != 0 ) ;
            assert ( cfg_vertex . count ( bb_target ) == 1 ) ;
            new CFG_Edge ( uid , cv_end , 
                    cfg_vertex [ bb_target ] , frequency ) ;
            uid ++ ;

            if ( bb -> branch_type == BT_Unconditional )
            {
                bb -> branch_type = BT_None ;
            }
        }
        // At end of basic block - branches handled
        if ( branch_links . count ( bb_end ) != 0 )
        {
            for ( k = branch_links [ bb_end ] . begin () ; 
                    k != branch_links [ bb_end ] . end () ; k ++ )
            {
                ulongint bb_target = (* k) . first ;
                ulongint frequency = (* k) . second ;

                assert ( cfg_vertex . count ( bb_target ) == 1 ) ;
                CFG_Edge * e = new CFG_Edge ( uid , cv_end , 
                        cfg_vertex [ bb_target ] , frequency ) ;
                e -> branch_cost = BRANCH_COST ;
                uid ++ ;
            }
        }
        bb_start = (* i) ;
    }
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;

        cv -> Check_Frequency () ;
    }
    // Step 3: Trim redundant vertexes
    bool changes_made = true ;
    while ( changes_made )
    {
        changes_made = false ;
        for ( l = all_cfg_vertexes . begin () ;
                l != all_cfg_vertexes . end () ; )
        {
            CFG_Vertex * cv = (* l) ;

            if ( cv -> Trim () )
            {
                delete cv ;
                changes_made = true ;
                l = all_cfg_vertexes . erase ( l ) ;
            } else {
                l ++ ;
            }
        }
    }
    // Step 4: Check frequencies and make list of edges
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;

        cv -> Check_Frequency () ;
        cv -> Add_Edges ( all_cfg_edges ) ;
        if ( cv -> Is_Entrance () )
        {
            assert ( entrance == 0L );
            entrance = cv ;
        }
    }
    assert ( entrance != 0L );

    // Step 5: Check instructions are unique
    std::unordered_set<ulongint> known ;
    std::list<CFG_Edge *>::iterator m ;

    for ( m = all_cfg_edges . begin () ;
            m != all_cfg_edges . end () ; m ++ )
    {
        CFG_Edge * ce = (* m) ;
        ce -> Check_Instruction ( known ) ;
    }
}

void Make_CFG :: Unmake ()
{
    std::list<CFG_Edge *>::iterator i ;
    std::list<CFG_Vertex *>::iterator l ;

    for ( i = all_cfg_edges . begin () ;
            i != all_cfg_edges . end () ; i ++ )
    {
        CFG_Edge * ce = (* i) ;
        delete ce ;
    }
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;
        delete cv ;
    }
    std::list<Base_Creator *>::iterator k ;
    
    for ( k = all_base_creators . begin () ;
            k != all_base_creators . end () ; k ++ )
    {
        Base_Creator * bc = (* k);
        delete bc;
    }
    all_base_creators . clear () ;
    all_cfg_vertexes . clear () ;
    all_cfg_edges . clear () ;
    entrance = 0L ;
}

bool Make_CFG :: Hacked_Up_Def_Use_Analysis ( FILE * ftr0fd ,
            FILE * ftr1fd , ulongint main_start , ulongint main_end )
{
    ulongint                            current_pc = 0 ;
    std::map<ulongint, CFG_Edge *>      bb_lookup;
    std::list<CFG_Edge *>::iterator     j;
    CFG_Vertex *                        vertex;
    CFG_Edge *                          cur_basic_block = 0L;
    struct FastTraceRecord              ftrp [ 2 ];
    struct FastTraceRecord *            ftr0 = & ftrp [ 0 ];
    struct FastTraceRecord *            ftr1 = & ftrp [ 1 ];
    ulongint                            i;
    Base_Instance *                     base_pointer = 0L;
    bool                                about_to_activate = false ;
    bool                                active = false ;
    bool                                can_activate = true ;
    ulongint                            current_sp = 0;
    ulongint                            top_of_stack = 0;
    ulongint                            counter = 0;
    ulongint                            current_fp = 0;
    ulongint                            previous_fp = 0;
    std::vector<ulongint>               fp_stack;
    std::vector<Base_Instance *>        bp_in_register ;
    std::vector<Base_Instance *>        bp_in_stack ;

    for(j = all_cfg_edges . begin (); 
                j != all_cfg_edges . end (); j++) {

        CFG_Edge * ce = (* j);

        if (ce->path.size() != 0) {
            ulongint pc = ce->Start_PC();
            bb_lookup[ pc ] = ce ;
        }
    }
    bp_in_register.resize(RFILE_SIZE, 0L);
    bp_in_stack.resize(32, 0L);
    uid ++;

    while (( fread ( ftr0 , 
            sizeof ( struct FastTraceRecord ) , 1 , ftr0fd ) == 1 )
    && ( fread ( ftr1 , 
            sizeof ( struct FastTraceRecord ) , 1 , ftr1fd ) == 1 )) {

        // assert synchronised
        if ( ftr0 -> pc_value != ftr1 -> pc_value ) {
            fprintf(stderr, "Warning: Executions desynced\n");
            fprintf(stderr, "stream 0: pc %lx\n", ftr0 -> pc_value);
            fprintf(stderr, "stream 1: pc %lx\n", ftr1 -> pc_value);
            return false;
        }
        counter ++;
        assert ( ftr0 -> optype == ftr1 -> optype ) ;
        assert ( ftr0 -> out_reg_number == ftr1 -> out_reg_number ) ;
        assert ( ftr0 -> in0_reg_number == ftr1 -> in0_reg_number ) ;
        assert ( ftr0 -> in1_reg_number == ftr1 -> in1_reg_number ) ;
        assert ( ftr0 -> store_reg_number == ftr1 -> store_reg_number ) ;

        if (ftr0 -> optype == 'T')
        {
            current_pc += 4;
        } else {
            current_pc = ftr0 -> pc_value;
        }
        if (ftr0->out_reg_number == SP_REGISTER) {
            if (top_of_stack == 0L) {
                current_sp = top_of_stack = ftr0->out_value;
                current_fp = previous_fp = top_of_stack;
                fprintf(stderr, "Stack pointer set to %"FX" at %"FX"\n",
                    top_of_stack,
                    current_pc);
            }
        }
        if ((counter & COUNT_INTERVAL) == 0) {
            fprintf (stderr, "HUDUA %"FU" %s\n", counter, 
                active ? "active" : 
                    ( can_activate ? "before" : "after" ));
        }


        // Activation mechanism - activation begins when
        // "main" calls "wcetmain", and ends when "wcetmain" returns.
        bool in_main = (( main_start <= current_pc )
            && ( current_pc < main_end )) ;
        if (active) {
            if (in_main) {
                // deactivate
                active = false;
            }
            about_to_activate = false ;
            can_activate = false ;
        } else {
            if (in_main) {
                about_to_activate = can_activate ;
            } else if (about_to_activate) {
                active = true ;
            }
        }

        if (! active) {
            continue ;
        }


        bool first_instruction = false;

        if (bb_lookup.count (current_pc) != 0) 
        {
            /* New basic block */
            cur_basic_block = bb_lookup[ current_pc ];
            vertex = cur_basic_block->start;
            first_instruction = true;
        }
        /* fprintf(stderr, "%"FX"   %c %d %d %010"FX"\n",
                current_pc, ftr0->optype, 
                (ftr0->out_value != ftr1->out_value),
                (ftr0->ea_value != ftr1->ea_value),
                ftr0->out_value) ; */
        assert (cur_basic_block != 0L);
        assert (cur_basic_block->Start_PC() <= current_pc);
        assert (current_pc <= cur_basic_block->End_PC());
        
        /* During transition between basic blocks, the vertex carries 
         * the list of active base pointers: */
        if (first_instruction) {
            ulongint i;
            for (i = 0; i < bp_in_register.size(); i ++) {
                if (bp_in_register[ i ] != 0L) {
                    assert (bp_in_register[ i ]->ref_count > 0);
                    vertex->Present(bp_in_register[ i ]);
                }
            }
            for (i = 0; i < bp_in_stack.size(); i ++) {
                if (bp_in_stack[ i ] != 0L) {
                    assert (bp_in_stack[ i ]->ref_count > 0);
                    vertex->Present(bp_in_stack[ i ]);
                }
            }
        }

        bool is_load = (ftr0 -> optype == 'L' );
        bool is_store = (ftr0 -> optype == 'S' );

        /* Identify the input and output pointers */
        Base_Instance * in0_pointer = 
                    bp_in_register[ftr0 -> in0_reg_number];
        Base_Instance * in1_pointer = 
                    bp_in_register[ftr0 -> in1_reg_number];
        Base_Instance * store_base_pointer = 
                    bp_in_register[ftr0 -> store_reg_number];

        if ((in0_pointer != 0L) && (in1_pointer != 0L)) {
            base_pointer = 0L ;
        } else if (in0_pointer != 0L) {
            base_pointer = in0_pointer ;
        } else if (in1_pointer != 0L) {
            base_pointer = in1_pointer ;
        } else {
            base_pointer = 0L ;
        }
        if (is_load && (ftr0 -> in0_reg_number == GP_REGISTER) // gp (on Alpha)
        && SUPPORT_LITERALS)
        {
            // That's no ordinary load! That's an access to the literal
            // table, and must be treated accordingly. Specifically, it
            // must be treated as if an immediate has been copied into
            // a register.
            is_load = false;
            base_pointer = 0L;
        }

        // How is a pointer used here?
        if (is_load || is_store) {
            if ( base_pointer == 0L ) {
                // A static pointer, used for load/store
                if ( ftr0 -> ea_value == ftr1 -> ea_value ) {
                    // (1) Because it's right at the top of the stack,
                    // in the argv/environment area
                    // or (2) Because it's in the text segment,
                    // e.g. switch statement offset 
                } else {
                    /* There's no good reason for this being static! */
                    fprintf(stderr, "pointer should not be static at pc %lx\n", 
                                    ftr0 -> pc_value ) ;
                    fprintf(stderr, "ea(1) is %lx\n", ftr0 -> ea_value ) ;
                    fprintf(stderr, "ea(2) is %lx\n", ftr1 -> ea_value ) ;
                    fprintf(stderr, "in0 %p in1 %p store %p\n" ,
                                    in0_pointer , in1_pointer , 
                                    store_base_pointer ) ;
                    return false;
                }
            } else {
                // It's a dynamic pointer, used for load/store
                cur_basic_block->Use(base_pointer, 
                            current_pc, ftr0->ea_value);

                // is this a stack-writing operation?
                if (base_pointer->Is_SP() && is_store) {
                    assert (top_of_stack != 0L);
                    int stack_offset = top_of_stack - ftr0->ea_value;
                    
                    if ((stack_offset < (int) (-MAX_STACK_OFFSET))
                    || (stack_offset > (int) MAX_STACK_OFFSET)) {
                        fprintf(stderr, "store: "
                            "improbable stack offset %d at pc %lx\n", 
                                        stack_offset, ftr0 -> pc_value ) ;
                        fprintf(stderr, "ea is %lx\n", ftr0 -> ea_value);
                        return false;
                    }
                    stack_offset /= STACK_ITEM_SIZE ;
                    if (bp_in_stack.size () <= (ulongint) stack_offset) {
                        bp_in_stack.resize(stack_offset + 32, 0L);
                    }
                        
                    if (bp_in_stack[ stack_offset ] != 0L) {
                        BP_Destroy(bp_in_stack[ stack_offset ]);
                        bp_in_stack[ stack_offset ] = 0L;
                    }
                    if (store_base_pointer != 0L) {
                        // Make new occupant
                        store_base_pointer->Inc_Ref_Count();
                        bp_in_stack[ stack_offset ] = store_base_pointer;
                    }
                }
            }
        }

        // Detect pointers by looking
        // at the output value - this will only be different between the
        // two simulators if a pointer is being processed by the instruction.
        bool is_pointer = ( ftr0 -> out_value != ftr1 -> out_value ) ;
        bool is_stack_pointer = false;

        if (ftr0->out_reg_number == SP_REGISTER) {
            // special case for stack pointer
            is_pointer = true;
            is_stack_pointer = true ;
            ulongint next_sp = ftr0 -> out_value ;

            assert (top_of_stack != 0L);
            if ( next_sp > current_sp ) {
                // Stack has got smaller 
                // zero the stack space that's no longer in use
                int32_t start_zero = top_of_stack - next_sp;
                int32_t end_zero = top_of_stack - current_sp;
                start_zero /= STACK_ITEM_SIZE ;
                end_zero /= STACK_ITEM_SIZE ;

                if ( start_zero >= 0 ) {
                    assert ( end_zero > start_zero ) ;
                    for (i = start_zero; (i < (ulongint) end_zero) 
                                && (i < bp_in_stack.size()); i++) {
                        if (bp_in_stack[ i ] != 0L) {
                            BP_Destroy(bp_in_stack[ i ]);
                            bp_in_stack [ i ] = 0L;
                        }
                    }
                }
            }
            current_sp = next_sp ;
        }

#if 0
        if (ftr0->out_reg_number == FP_REGISTER) {
            // Change in FP; valid area of stack has changed
            ulongint fp = ftr0 -> out_value ;

            if (fp < current_fp) {
                // Call function
                //printf ("%"FX" called, %"FX", %"FU"\n", 
                //        ftr0->pc_value, fp, (ulongint) fp_stack.size());
                fp_stack.push_back(current_fp);
                current_fp = fp;
            } else {
                // Return
                //printf ("%"FX" return, %"FX", %"FU"\n", 
                //        ftr0->pc_value, fp, (ulongint) fp_stack.size());
                if (!fp_stack.empty()) {
                    ulongint fp0 = fp_stack.back();
                    if (fp0 != top_of_stack) {
                        assert (fp == fp0);
                    }
                    fp_stack.pop_back();
                }
                current_fp = fp;
            }
            if (fp_stack.size() > 1) {
                previous_fp = fp_stack[ fp_stack.size() - 2 ];
            } else {
                previous_fp = top_of_stack;
            }
            assert (current_fp < previous_fp);
        }
#endif

        // If this operation creates or copies a base pointer,
        // update the base pointer table. The purpose of the following
        // code is to choose values for create, is_fixed, copy, destroy:
        bool create = false;
        bool is_fixed = false;
        Base_Instance * copy = 0L;
        bool destroy = true;

        if ((ftr0 -> out_reg_number != NOT_A_REGISTER)
        && (ftr0 -> out_reg_number != ZERO_REGISTER)
        && (is_pointer)) {
            if (is_load) {
                // Base pointer created dynamically by load
                if ((base_pointer != 0L)
                && ((ftr0 -> in0_reg_number == FP_REGISTER)
                    || (ftr0 -> in0_reg_number == SP_REGISTER))
                && (base_pointer->Is_SP())) {
                    assert (top_of_stack != 0);

                    // Load from stack - do we know which pointer this is?
                    int stack_offset = top_of_stack - ftr0->ea_value;
                    if ((stack_offset < 0)
                    || (stack_offset > (int) MAX_STACK_OFFSET))
                    {
                        fprintf(stderr, "load: "
                            "improbable stack offset %d at pc %lx\n", 
                                        stack_offset, ftr0 -> pc_value ) ;
                        fprintf(stderr, "ea is %lx\n", ftr0 -> ea_value);
                        return false;
                    }
                    stack_offset /= STACK_ITEM_SIZE ;
                    if (((ulongint) stack_offset < bp_in_stack . size ())
                    && (bp_in_stack[ stack_offset ] != 0L)) {
                        // Yes, it's one we saved earlier, copy it
                        copy = bp_in_stack[ stack_offset ] ;
                    } else {
                        // No, must be a new pointer
                        create = true ;
                    }
                } else {
                    // Not from stack - must be a new pointer
                    create = true;
                }
            } else if (base_pointer != 0L) {
                // Base pointer copied
                if (bp_in_register[ ftr0 -> out_reg_number ] 
                                == base_pointer) {
                    // Copy to same register -> no change
                    destroy = false ;
                } else {
                    copy = base_pointer ;
                }
            } else {
                // Base pointer created statically -
                // a fixed memory location.
                create = true ;
                is_fixed = true ;
            }
        } else {
            // Output register does not hold a pointer
        }

        // Now apply the commands decided in the previous section
        // create, destroy, is_fixed, copy.
        if (destroy) {
            // Destroy previous occupant of the output register
            Base_Instance * bi = bp_in_register[ ftr0 -> out_reg_number ];
            if (bi != 0L) {
                BP_Destroy(bi);
                bp_in_register[ ftr0 -> out_reg_number ] = 0L;
            }
        }
        if (copy != 0L) {
            // Copy to new register
            copy->Inc_Ref_Count();
            bp_in_register[ ftr0 -> out_reg_number ] = copy;
        }
        if (create) {
            // Create a new base pointer
            base_pointer = cur_basic_block->Define(current_pc, uid,
                        ftr0->out_value, is_stack_pointer);
            uid ++;
            bp_in_register[ ftr0->out_reg_number ] = base_pointer ;
        } 
    }
    assert (top_of_stack != 0L);
    for (i = 0; i < bp_in_stack.size(); i++) {
        if (bp_in_stack[ i ] != 0L) {
            BP_Destroy(bp_in_stack[ i ]);
        }
    }
    for (i = 0; i < bp_in_register.size(); i++) {
        if (bp_in_register[ i ] != 0L) {
            BP_Destroy(bp_in_register[ i ]);
        }
    }

    for(j = all_cfg_edges . begin (); 
                j != all_cfg_edges . end (); j++) {

        CFG_Edge * ce = (* j);
        std::map<ulongint, Base_Creator *>::iterator k;

        for (k = ce->creator.begin();
                    k != ce->creator.end(); k++) {
            Base_Creator * bc = (* k).second;
            assert(bc->active_instances.size() == 0);
            all_base_creators.push_back(bc);
        }
    }
    fprintf(stderr, "number of base creators: %"FU"\n",
            all_base_creators.size());
    return true;
}


void Make_CFG :: Export ( FILE * outfd )
{
    fprintf ( outfd , "exf %d %"FU"\n", (int) early_exit_flag , inst_count ) ;

    std::list<CFG_Vertex *>::iterator i ;
    
    for ( i = all_cfg_vertexes . begin () ;
            i != all_cfg_vertexes . end () ; i ++ )
    {
        (* i) -> Export ( outfd ) ;
    }

    std::list<CFG_Edge *>::iterator j ;
    
    for ( j = all_cfg_edges . begin () ;
            j != all_cfg_edges . end () ; j ++ )
    {
        (* j) -> Export ( outfd , mem_access ) ;
    }

    std::list<Base_Creator *>::iterator k ;
    
    for ( k = all_base_creators . begin () ;
            k != all_base_creators . end () ; k ++ )
    {
        (* k) -> Export ( outfd ) ;
    }
}

void Make_CFG :: Export_UDG ( FILE * udgf , Symbol_Table * symtab )
{
    std::list<CFG_Vertex *>::iterator i ;
    
    for ( i = all_cfg_vertexes . begin () ;
            i != all_cfg_vertexes . end () ; i ++ )
    {
        (* i) -> udg_printed = false;
    }

    std::list<Base_Creator *>::iterator k ;
    
    for ( k = all_base_creators . begin () ;
            k != all_base_creators . end () ; k ++ )
    {
        (* k) -> udg_printed = false;
    }
    fputc ( '[' , udgf ) ;
    entrance -> Export_UDG ( udgf , symtab ) ;
    fprintf ( udgf , "]\n" ) ;
}


CFG_Vertex :: CFG_Vertex ( ulongint id , ulongint pc )
{
    this -> id = id ;
    this -> pc = pc ;
    udg_printed = false ;
}

bool CFG_Vertex :: Trim ()
{
    // Return true if this CFG_Vertex should be deleted
    if (( entrances . size () == 0 )
    && ( exits . size () == 0 ))
    {
        // I'm disconnected!
        return true ;
    } else if ( entrances . size () == 0 )
    {
        // I'm the program start node
        return false ;
    } else if ( exits . size () == 0 )
    {
        // I'm the program exit node
        return false ;
    } else if (( entrances . size () == 1 )
    && ( exits . size () == 1 ))
    {
        // I might be redundant.
        //  [1] ---- [2] ---- [3]   becomes [1] ---- [3]

        CFG_Vertex * cv2 = this ;
        CFG_Edge * ce12 = (* (entrances . begin ()));
        CFG_Edge * ce23 = (* (exits . begin ()));
        //CFG_Vertex * cv1 = ce12 -> start ;
        CFG_Vertex * cv3 = ce23 -> end ;

        // Sanity check
        assert ( ce12 -> end == cv2 ) ;
        assert ( ce23 -> start == cv2 ) ;

        if ( ce12 -> branch_type == BT_None )
        {
            if (( ce12 -> path . size () == 0 )
            || ( ce23 -> path . size () == 0 ))
            {
                // Merge ce12 and ce23, trim cv2 
                if ( ce12 -> Merge ( ce23 ) )
                {
                    ce12 -> end = cv3 ;
                    ce12 -> branch_type = ce23 -> branch_type ;
                    cv3 -> entrances . erase ( ce23 ) ;
                    cv3 -> entrances . insert ( ce12 ) ;
                    delete ce23 ;
                    return true ;
                }
            }
        }
    }
    return false ;
}

void CFG_Vertex :: Export ( FILE * outfd )
{
    CFG_Edge_Set::iterator i;

    fprintf ( outfd , "v %06"FX"\n" , id ) ;
    if ( entrances . size () == 0 )
    {
        fprintf ( outfd , "  entrance\n" ) ;
    }
    if ( exits . size () == 0 )
    {
        fprintf ( outfd , "  exit\n" ) ;
    }
    if ( pc != 0 )
    {
        fprintf ( outfd , "  pc %"FX"\n" , pc ) ;
    }
    for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
    {
        fprintf ( outfd , "  ie %06"FX"\n" , (* i) -> id ) ;
    }
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        fprintf ( outfd , "  oe %06"FX"\n" , (* i) -> id ) ;
    }
        
    std::map<ulongint, Base_Creator *>::iterator j;

    for (j = present.begin(); j != present.end(); j++) {
        fprintf ( outfd , "  present %06"FX"\n" , (* j).second->id ) ;
    }
}

void CFG_Vertex :: Check_Frequency ()
{
    CFG_Edge_Set::iterator  i;
    ulongint                total_in = 0 ;
    ulongint                total_out = 0 ;
    ulongint                in_zero_count = 0 ;
    ulongint                out_zero_count = 0 ;
    ulongint                diff ;
    static int jatt = 0;

    jatt ++;

    for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
    {
        total_in += (* i) -> frequency ;
        if ( (* i) -> frequency == 0 )
        {
            in_zero_count ++ ;
        }
    }
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        total_out += (* i) -> frequency ;
        if ( (* i) -> frequency == 0 )
        {
            out_zero_count ++ ;
        }
    }
    assert (( in_zero_count + out_zero_count ) <= 1 ) ;
    if ( in_zero_count == 1 )
    {
        assert ( total_out > total_in ) ;
        diff = total_out - total_in ;
        for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
        {
            if ( (* i) -> frequency == 0 )
            {
                (* i) -> frequency = diff ;
                break ;
            }
        }
    } else if ( out_zero_count == 1 )
    {
        assert ( total_in > total_out ) ;
        diff = total_in - total_out ;
        for ( i = exits . begin () ; i != exits . end () ; i ++ )
        {
            if ( (* i) -> frequency == 0 )
            {
                (* i) -> frequency = diff ;
                break ;
            }
        }
    } else {
        assert (( total_in == total_out ) || Is_Entrance () || Is_Exit () ) ;
    }
}

void CFG_Vertex :: Add_Edges ( std::list<CFG_Edge *> & all_edges )
{
    CFG_Edge_Set::iterator  i;

    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        all_edges . push_back ( (* i) ) ;
    }
}

void CFG_Vertex :: Export_UDG ( FILE * udgf , Symbol_Table * symtab )
{
    if ( udg_printed )
    {
        fprintf ( udgf , "r(\"L%"FX"\")" , id ) ;
        return ;
    }
    udg_printed = true ;
    fprintf ( udgf , "l(\"L%"FX"\",n(\"C%"FU"\",[a(\"OBJECT\",\" %"FU , 
                id , id , id ) ;
    if ( pc == 0 )
    {
        //fprintf ( udgf , "" ) ;
    } else if ( symtab == 0L )
    {
        fprintf ( udgf , "%"FX"" , pc ) ;
    } else {
        symtab -> Lookup ( udgf , pc ) ;
    }
    fprintf ( udgf , "\")],[" ) ;

    CFG_Edge_Set::iterator i ;
    bool first = true ;
    
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        CFG_Edge * e = (* i) ;

        if ( ! first ) 
            fprintf ( udgf , "," ) ;

        first = false ;
        fprintf ( udgf , "e(\"E%"FX"\",[],"
            "l(\"L%"FX"x\",n(\"C%"FU"x\","
            "[a(\"OBJECT\",\"", e -> id, e->id, e->id ) ;

        /* edge label */
        std::vector<std::pair<ulongint, ulongint> >::iterator k ;

        fprintf ( udgf , "%"FU, e -> frequency );
        k = e->path.begin ();
        if (e->path.size() > 7)
        {
            // Print some
            ulongint j = 0;
            for ( j = 0 ; j < 4 ; j ++ )
            {
                fprintf ( udgf, "\\n%"FX"" , (* k) . first ) ;
                k ++;
            }
            fprintf ( udgf, "\\n  ..." );
            k = e->path.end();
            k --; k --; k --;
        }
        // Print all
        for ( ; k != e->path . end () ; k ++ )
        {
            fprintf ( udgf, "\\n%"FX"" , (* k) . first ) ;
        }

        fprintf ( udgf , "\")],[e(\"E%"FX"x\",[],", e->id ) ;
        e -> end -> Export_UDG ( udgf , symtab ) ;

        std::map<ulongint, Base_Creator *>::iterator j;
        for (j = e->creator.begin(); j != e->creator.end(); j++) {
            if ((* j).second->Is_Important()) {
                fprintf ( udgf , "),e(\"CR%"FX"\",["
                    BASE_EDGE_ATTR
                    "a(\"OBJECT\",\"creates\")],", (* j).first);
                (* j).second->Export_UDG(udgf);
            }
        }
        std::map<ulongint, Usage_Data>::iterator l;
        for (l = e->user.begin(); l != e->user.end(); l++) {
            if ((* l).second.source->Is_Important()) {
                fprintf ( udgf , "),e(\"US%"FX"\",["
                    "a(\"_DIR\",\"first\"),"
                    BASE_EDGE_ATTR
                    "a(\"OBJECT\",\"uses x%"FU"\")],", (* l).first,
                        (* l).second.used_at.size());
                (* l).second.source->Export_UDG(udgf);
            }
        }

        fprintf ( udgf , ")])))" ) ;
    }
    fprintf ( udgf , "]))" ) ;
}    

CFG_Edge :: CFG_Edge ( ulongint id , CFG_Vertex * start ,
                CFG_Vertex * end , ulongint frequency )
{
    this -> id = id ;
    this -> start = start ;
    this -> end = end ;
    this -> frequency = frequency ;
    this -> start -> exits . insert ( this ) ;
    this -> end -> entrances . insert ( this ) ;
    this -> branch_type = BT_None ;
    this -> branch_cost = 0 ;
}

void CFG_Edge :: Export ( FILE * outfd ,
        std::unordered_set<ulongint> & mem_access )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    fprintf ( outfd , "e %06"FX"\n" , id ) ;
    fprintf ( outfd , "  f %"FX"\n" , frequency ) ;
    fprintf ( outfd , "  iv %06"FX"\n" , start -> id ) ;
    fprintf ( outfd , "  ov %06"FX"\n" , end -> id ) ;
    fprintf ( outfd , "  bt " ) ;
    switch ( branch_type )
    {
        case BT_None :          fprintf ( outfd , "None" ) ; break ;
        case BT_Unconditional : fprintf ( outfd , "Unconditional" ) ; break ;
        case BT_Conditional :   fprintf ( outfd , "Conditional" ) ; break ;
        case BT_Serialising :   fprintf ( outfd , "Serialising" ) ; break ;
        default :
            assert ( 0 ) ;
            break ;
    }
    fprintf ( outfd , "\n" ) ;

    ulongint cost = branch_cost ;
    for ( i = path . begin () ; i != path . end () ; i ++ )
    {
        fprintf ( outfd , "  x %"FX"%s\n" , (* i) . first ,
            ( mem_access . count ( (* i) . first ) != 0 ) ? " mem" : "" );
        cost += (* i) . second;
    }
    fprintf ( outfd , "  cost %"FX"\n" , cost ) ;

    std::map<ulongint, Base_Creator *>::iterator j;
    for (j = creator.begin(); j != creator.end(); j++) {
        fprintf ( outfd , "  creator %06"FX"\n" , (* j).second->id ) ;
    }

    std::map<ulongint, Usage_Data>::iterator l;
    for (l = user.begin(); l != user.end(); l++) {
        std::set<ulongint>::iterator m;
        
        fprintf ( outfd , "  user %06"FX,
                        (* l).second.source->id);
        for (m = (* l).second.used_at.begin();
                m != (* l).second.used_at.end(); m++) {
            fprintf(outfd , " %"FX, (* m));
        }
        fprintf(outfd, "\n");
    }
}

void CFG_Edge :: Check_Instruction ( std::unordered_set<ulongint> & known )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    for ( i = path . begin () ; i != path . end () ; i ++ )
    {
        assert ( known . count ( (* i) . first ) == 0 ) ;
        known . insert ( (* i) . first ) ;
    }
}

void CFG_Edge :: Add_Instruction ( ulongint pc , 
                ulongint repeats , ulongint cost )
{
    assert ( repeats != 0 );
    assert ( cost != 0 );
    cost *= repeats;
    this -> path . push_back ( 
            std::pair<ulongint, ulongint> ( pc , cost ) ) ;
}

bool CFG_Edge :: Merge ( CFG_Edge * other )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    if ( other -> frequency != 0 )
    {
        if ( this -> frequency != 0 )
        {
            assert ( this -> frequency == other -> frequency ) ;
        } else {
            this -> frequency = other -> frequency ;
        }
    }
    for ( i = other -> path . begin () ; i != other -> path . end () ; i ++ )
    {
        this -> Add_Instruction ( (* i) . first , 1 , (* i) . second ) ;
    }
    return true ; // Merge successful
}

Base_Creator :: Base_Creator(ulongint _pc, 
            ulongint _id, bool _is_stack_pointer)
{
    pc = _pc;
    id = _id;
    is_stack_pointer = _is_stack_pointer;
    upper_bound = INT_MIN;
    lower_bound = INT_MAX;
    max_instances = 0;
    active_instances.clear();
    inactive_instances.clear();
    udg_printed = false;
}





Base_Instance :: Base_Instance(Base_Creator * _source, ulongint _initial_value)
{
    // New instance of a pointer created by Base_Creator enters the system
    source = _source;
    initial_value = _initial_value;

    // This instance is added to the collection controlled by Base_Creator
    if (source->inactive_instances.size() == 0) {
        instance_number = source->inactive_instances.size() + 
            source->active_instances.size();
    } else {
        std::set<ulongint>::iterator i;

        i = source->inactive_instances.begin();
        instance_number = (* i);
        source->inactive_instances.erase(i);
    }
    assert (source->active_instances.count(instance_number) == 0);
    assert (source->inactive_instances.count(instance_number) == 0);
    source->active_instances.insert(instance_number);

    if (source->active_instances.size() > source->max_instances) {
        source->max_instances = source->active_instances.size();
    }
    ref_count = 1;
}

Base_Instance :: ~Base_Instance()
{
    // Instance of a pointer is removed from the system
    assert (source->active_instances.count(instance_number) == 1);
    assert (source->inactive_instances.count(instance_number) == 0);
    source->active_instances.erase(instance_number);
    source->inactive_instances.insert(instance_number);
}

void Base_Instance :: Use(ulongint address, ulongint pc)
{
    // A pointer is used by an instruction at "pc"
    Base_Creator *  bc = source;
    int32_t         offset = (int32_t) ( address - initial_value );

    if (offset < bc->lower_bound) {
        bc->lower_bound = offset;
    }
    offset += 8;
    if (offset > bc->upper_bound) {
        bc->upper_bound = offset;
    }

    // What sort of base pointer is this? Can we classify its
    // access pattern?
    if (bc->ap_last_access.count(pc) != 0) {

        Access_Data & ad = bc->ap_last_access[ pc ];
        int step = offset - ad.last_offset;

        switch (ad.ap_classification) {
            case AP_NOT_ITERATIVE:
            case AP_PSEUDO_STATIC:
                if (step != 0) {
                    ad.ap_constant_step = step;
                    ad.ap_classification = AP_CONSTANT_STEP;
                }
                break;
            case AP_CONSTANT_STEP:
            case AP_VARIABLE_STEP:
                if (step != ad.ap_constant_step) {
                    // Counterexample found.
                    if (((step > 0) && (ad.ap_constant_step > 0))
                    || ((step <= 0) && (ad.ap_constant_step <= 0))) {
                        // Variable stepping (in same direction)
                        ad.ap_classification = AP_VARIABLE_STEP;
                    } else if (offset == ad.initial_offset) {
                        // Special case: rewind to start of array
                    } else {
                        // Otherwise: random
                        ad.ap_classification = AP_RANDOM_ACCESS;
                    }
                }
                break;
            case AP_RANDOM_ACCESS:
                break;
        }
        ad.counter ++;
        ad.sample = address;
        ad.last_offset = offset;
    } else {
        Access_Data ad;
        ad.counter = 1;
        ad.ap_classification = AP_NOT_ITERATIVE;
        ad.sample = address;
        ad.ap_constant_step = 0;
        ad.last_offset = offset;
        ad.initial_offset = offset;
        bc->ap_last_access[ pc ] = ad;
    }
}

int32_t Base_Instance :: Stack_Offset(ulongint ea_value)
{
    assert (Is_SP());
    return (int32_t) (initial_value - ea_value);
}

Base_Instance * CFG_Edge :: Define(ulongint _pc, ulongint _id,
            ulongint _initial_value, bool _is_stack_pointer)
{
    // New instance. Also make a Base_Creator if this instruction
    // hasn't been executed before.
    if (creator.count(_pc) == 0) {
        assert ( _pc >= Start_PC() );
        assert ( _pc <= End_PC() );
        creator[ _pc ] = new Base_Creator(_pc, _id, _is_stack_pointer);
    }
    return new Base_Instance(creator[ _pc ], _initial_value);
}

void CFG_Edge :: Use(Base_Instance * bi, ulongint pc, ulongint address)
{
    // A pointer is used by an edge
    bi->Use(address, pc);
    user[ bi->source->pc ].source = bi->source;
    user[ bi->source->pc ].used_at.insert(pc);
}

void CFG_Vertex :: Present(Base_Instance * bi)
{
    //if (present.count(bi->source->pc) == 0) {
    //    present[ bi->source->pc ] = bi->source;
    //} 
}

ulongint Base_Instance :: Dec_Ref_Count()
{ 
    assert (ref_count > 0);
    ref_count --; 
    return ref_count; 
}

void BP_Destroy(Base_Instance * bi)
{
    if (bi->Dec_Ref_Count() == 0) {
        delete bi;
    }
}

void Base_Creator :: Export ( FILE * outfd )
{

    fprintf ( outfd , "b %06"FX"\n" , id ) ;
    if (!Is_Valid() || (ap_last_access.size() == 0)) {
        fprintf ( outfd , "  type invalid\n");
    } else if (Is_SP()) {
        fprintf ( outfd , "  type stack\n");
    } else {
        assert (upper_bound != INT_MIN);
        assert (lower_bound != INT_MAX);
        fprintf ( outfd , "  type valid\n");
        fprintf ( outfd , "  max %"FX"\n" , max_instances ) ;
        fprintf ( outfd , "  size %"FX"\n" , 
            ((ulongint) upper_bound) - ((ulongint) lower_bound));
        fprintf ( outfd , "  sample %"FX"\n" , 
                        (*(ap_last_access.begin())).second.sample);

        std::map<ulongint, Access_Data>::iterator i;

        for (i = ap_last_access.begin(); 
                i != ap_last_access.end(); i ++) {

            Access_Data & ad = (* i).second;

            fprintf(outfd, "  class %"FX" ", (* i).first); // pc

            switch (ad.ap_classification) {
                case AP_NOT_ITERATIVE:
                    fprintf(outfd, "nit 0 ");
                    break;
                case AP_PSEUDO_STATIC:
                    fprintf(outfd, "pst 0 ");
                    break;
                case AP_CONSTANT_STEP:
                    fprintf(outfd, "cst %d ", ad.ap_constant_step);
                    break;
                case AP_VARIABLE_STEP:
                    fprintf(outfd, "vst %d ", ad.ap_constant_step);
                    break;
                case AP_RANDOM_ACCESS:
                    fprintf(outfd, "rnd 0 ");
                    break;
            }
            fprintf(outfd, "%"FX" %"FX"\n", ad.sample, ad.counter );
        }
    }
}

void Base_Creator :: Export_UDG ( FILE * udgf )
{
    if ( udg_printed )
    {
        fprintf ( udgf , "r(\"LBC%"FX"\")" , pc ) ;
        return ;
    }
    udg_printed = true ;

    fprintf ( udgf , "l(\"LBC%"FX"\",n(\"CBC%"FU"\",[a(\"OBJECT\",\"",
                pc , pc ) ;
    ulongint size = ((ulongint) upper_bound) - ((ulongint) lower_bound);
    fprintf ( udgf , "size %"FU" max %"FU, size, max_instances);
    fprintf ( udgf , "\")],[]))");
}

ulongint CFG_Edge :: Start_PC()
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;
    assert (path.size() != 0);
    i = path.begin();
    return (* i). first;
}

ulongint CFG_Edge :: End_PC()
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;
    assert (path.size() != 0);
    i = path.end();
    i --;
    return (* i). first;
}



